
#include <cstring>
#include <boost/cstdint.hpp>

#include "byte_ordering.hpp"

namespace marshall {

namespace {

byte_ordering_t determine_ordering(void)
{
    const uint32_t value = 0x01020304;
    const uint8_t little[] = { 0x04, 0x03, 0x02, 0x01 };
    const uint8_t big[] = { 0x01, 0x02, 0x03, 0x04 };

    if (std::memcmp(static_cast<const void*>(&value), static_cast<const void*>(little), sizeof(value)) == 0)
        return Little;
    if (std::memcmp(static_cast<const void*>(&value), static_cast<const void*>(big), sizeof(value)) == 0)
        return Big;

    return Unknown;
}

byte_ordering_t endianness(void)
{
    static byte_ordering_t ordering = determine_ordering();
    return ordering;
}

template <typename T>
T byteswap(const T*);

template <>
uint16_t byteswap<uint16_t>(const uint16_t* value)
{
    return (((*value & 0xff00) >> 8) | 
            ((*value & 0x00ff) << 8));
}

template <>
uint32_t byteswap<uint32_t>(const uint32_t* value)
{
    return (((*value & 0xff000000) >> 24) | 
            ((*value & 0x00ff0000) >> 8)  |
            ((*value & 0x0000ff00) << 8)  |
            ((*value & 0x000000ff) << 24));
}

template <>
uint64_t byteswap<uint64_t>(const uint64_t* value)
{
    return (((*value & 0xff00000000000000) >> 56) | 
            ((*value & 0x00ff000000000000) >> 40) |
            ((*value & 0x0000ff0000000000) >> 24) |
            ((*value & 0x000000ff00000000) >> 8)  |
            ((*value & 0x00000000ff000000) << 8)  |
            ((*value & 0x0000000000ff0000) << 24) |
            ((*value & 0x000000000000ff00) << 40) |
            ((*value & 0x00000000000000ff) << 56));
}

// Default implementation for to_order16 - do nothing
template <byte_ordering_t, byte_ordering_t>
uint16_t to_order16(const uint16_t* value) { return *value; }

// Little- to big-endian implementation (and vice-versa) - reverse byte order
template <>
uint16_t to_order16<Big, Little>(const uint16_t* value)
{
    return marshall::byteswap<uint16_t>(value);
}
template <>
uint16_t to_order16<Little, Big>(const uint16_t* value)
{
    return marshall::byteswap<uint16_t>(value);
}

template <byte_ordering_t, byte_ordering_t>
uint32_t to_order32(const uint32_t* value) { return *value; }

template <>
uint32_t to_order32<Big, Little>(const uint32_t* value)
{
    return marshall::byteswap<uint32_t>(value);
}
template <>
uint32_t to_order32<Little, Big>(const uint32_t* value)
{
    return marshall::byteswap<uint32_t>(value);
}

#ifndef BOOST_NO_INT64_T
template <byte_ordering_t, byte_ordering_t>
uint64_t to_order64(const uint64_t* value) { return *value; }

template <>
uint64_t to_order64<Big, Little>(const uint64_t* value)
{
    return marshall::byteswap<uint64_t>(value);
}
template <>
uint64_t to_order64<Little, Big>(const uint64_t* value)
{
    return marshall::byteswap<uint64_t>(value);
}
#endif

} // anonymous namespace

// Function yielding the conversion function for 16-bit conversions
template <>
conversion_function_t conversion_function<uint16_t, Big>(void)
{
    conversion_function16_t big = &to_order16<Big, Big>;
    conversion_function16_t little = &to_order16<Big, Little>;
    return reinterpret_cast<conversion_function_t>(endianness() == Big ? big : little);
}
template <>
conversion_function_t conversion_function<uint16_t, Little>(void)
{
    conversion_function16_t big = &to_order16<Little, Big>;
    conversion_function16_t little = &to_order16<Little, Little>;
    return reinterpret_cast<conversion_function_t>(endianness() == Big ? big : little);
}

template <>
conversion_function_t conversion_function<uint32_t, Big>(void)
{
    conversion_function32_t big = &to_order32<Big, Big>;
    conversion_function32_t little = &to_order32<Big, Little>;
    return reinterpret_cast<conversion_function_t>(endianness() == Big ? big : little);
}
template <>
conversion_function_t conversion_function<uint32_t, Little>(void)
{
    conversion_function32_t big = &to_order32<Little, Big>;
    conversion_function32_t little = &to_order32<Little, Little>;
    return reinterpret_cast<conversion_function_t>(endianness() == Big ? big : little);
}

#ifndef BOOST_NO_INT64_T
template <>
conversion_function_t conversion_function<uint64_t, Big>(void)
{
    conversion_function64_t big = &to_order64<Big, Big>;
    conversion_function64_t little = &to_order64<Big, Little>;
    return reinterpret_cast<conversion_function_t>(endianness() == Big ? big : little);
}
template <>
conversion_function_t conversion_function<uint64_t, Little>(void)
{
    conversion_function64_t big = &to_order64<Little, Big>;
    conversion_function64_t little = &to_order64<Little, Little>;
    return reinterpret_cast<conversion_function_t>(endianness() == Big ? big : little);
}
#endif
} // namespace marshall

