#ifndef ORDERED_OARCHIVE_HPP
#define ORDERED_OARCHIVE_HPP

#include <ostream>
#include <string>
#include <vector>

#include <boost/utility/enable_if.hpp>

#include <boost/archive/basic_binary_oprimitive.hpp>
#include <boost/serialization/split_free.hpp>

#include "byte_ordering.hpp"
#include "unidentified_archive.hpp"

#include "ordered_types.hpp"

namespace boost { 
namespace archive {

using marshall::byte_ordering_t;
using marshall::conversion_function;

// The signature for a function to save a particular size in a particular ordering
template<size_t N, byte_ordering_t Order, class Primitive>
void ordered_save(Primitive* primitive, const void* address);

// A function to handle the standard save method
template<typename StoreType, typename Converter, byte_ordering_t Order, class Primitive>
inline void do_ordered_save(Primitive* primitive, const void* address)
{
    static Converter converter = reinterpret_cast<Converter>(conversion_function<StoreType, Order>());

    StoreType value = (*converter)(reinterpret_cast<const StoreType*>(address));
    primitive->save_binary(&value, sizeof(StoreType));
}

// The binary primitive for ordered input
template<class Archive, class OStream, byte_ordering_t Order>
struct ordered_binary_oprimitive :
    public basic_binary_oprimitive<Archive, OStream>
{
    ordered_binary_oprimitive(OStream & os, bool no_codecvt) :
        basic_binary_oprimitive<Archive, OStream>(os, no_codecvt)
    {
    }

    // Override the save function to do byte-ordered saving of the types we handle
    template<typename T>
    void save(const T& t, 
              typename boost::enable_if<boost::mpl::contains<ordered_types, T>, T>::type* dummy = 0)
    {
        // When saving data, align it to required boundaries
        This()->pre_align(t);

        aligned_save<T>(t);

        This()->post_align(t);
    }

    // Ensure that a bool is encoded as 8-bit
    void save(const bool& b)
    {
        uint8_t value = (b ? 1 : 0);
        save(value);
    }

    // Assume that a string's length has been determined externally
    void save(const std::string &s)
    {
        This()->pre_align(*s.begin());

        save_binary(s.data(), s.size());

        This()->post_align(s);
    }
    #ifndef BOOST_NO_STD_WSTRING
    void save(const std::wstring &ws)
    {
        This()->pre_align(*ws.begin());

        for (std::wstring::const_iterator it = ws.begin(); it != ws.end(); ++it)
            aligned_save(*it);

        This()->post_align(ws);
    }
    #endif

protected:
    friend class basic_binary_oarchive<Archive>;
    
    // Once aligned, save data in the appropriate byte ordering
    template<typename T>
    void aligned_save(const T & t)
    {
        ordered_save<sizeof(T), Order>(this, & t);
    }

    // Default alignment prior to insertion - no padding
    template<typename T>
    void pre_align(const T&) {}

    // Default alignment after insertion - no padding
    template<typename T>
    void post_align(const T&) {}
};

// Combine the ordered primitive with an unidentified archive
template<class Archive, byte_ordering_t Order>
struct ordered_oarchive_impl : 
    public ordered_binary_oprimitive<Archive, std::ostream, Order>,
    public unidentified_oarchive<Archive>
{
    ordered_oarchive_impl(std::ostream & os, unsigned int flags = no_header|no_codecvt) :
        ordered_binary_oprimitive<Archive, std::ostream, Order>(
            os, 
            0 != (flags & no_codecvt))
    {
        if(0 == (flags & no_header)){
            unidentified_oarchive<Archive>::init();
            ordered_binary_oprimitive<Archive, std::ostream, Order>::init();
        }
    }
};

template <byte_ordering_t Order>
struct ordered_oarchive : 
    public ordered_oarchive_impl<ordered_oarchive<Order>, Order>
{
    ordered_oarchive(std::ostream & os, unsigned int flags = no_header|no_codecvt) :
        ordered_oarchive_impl<ordered_oarchive<Order>, Order>(os, flags)
    {}
};

} // namespace archive

namespace serialization {

// Provide default implementation for serializing vectors to network archives
template<class Archive, marshall::byte_ordering_t Order, class T, class A>
inline void save(boost::archive::ordered_oarchive_impl<Archive, Order> & ar, 
                 const std::vector<T, A> & v,
                 const unsigned version)
{
    for (typename std::vector<T, A>::const_iterator it = v.begin(); it != v.end(); ++it)
        ar << (*it);
}

} // namespace serialization
} // namespace boost

#endif // ORDERED_OARCHIVE_HPP
